from textual.binding import Binding

VIM_OPTION_LIST_NAVIGATE = [
    Binding("j", "cursor_down", "Down", show=False),
    Binding("k", "cursor_up", "Up", show=False),
    Binding("G", "last", "Last", show=False),
    Binding("ctrl+b", "page_down", "Page Down", show=False),
    Binding("ctrl+f", "page_up", "Page Up", show=False),
]
BOOKMARK = Binding("b", "bookmark", "Bookmark")
CLEAR_CACHE = Binding("ctrl+d", "purge_from_cache", "Purge from cache")
OPEN_GITHUB = Binding("ctrl+g", "open_github", "Open GitHub")
GITHUB_STATS = Binding("ctrl+s", "github_stats", "GitHub Stats")
BACK = Binding("backspace", "back", "Back", show=False)
LEFT_BACK = Binding("left", "back", "Back", show=False)



from tofuref.widgets.code_block_select import CodeBlockSelect
from tofuref.widgets.content_window import ContentWindow
from tofuref.widgets.logo import Logo
from tofuref.widgets.providers_option_list import ProvidersOptionList
from tofuref.widgets.resources_option_list import ResourcesOptionList
from tofuref.widgets.search_input import SearchInput
from tofuref.widgets.start_progress import StartProgress
from tofuref.widgets.status import Status

__all__ = ["CodeBlockSelect", "ContentWindow", "Logo", "ProvidersOptionList", "ResourcesOptionList", "SearchInput", "StartProgress", "Status"]



from textual.widgets import Static

from tofuref import __version__


class Logo(Static):
    DEFAULT_CSS = """
    Logo {
        column-span: 1;
        width: auto;
        background: $surface;
    }
    """

    def render(self):
        # Using Rich markup directly to style colors like the HTML version
        return (
            "[$accent]  ┌───────┐\n"
            "[$secondary]│[/$secondary]  [$primary]tofuref[/$primary]  [$secondary]│[/$secondary]\n"
            f"[$accent]  └─{__version__}─┘"
            # f"[$accent]  └───────┘"
        )



import json
import logging
import time
from collections.abc import Collection
from pathlib import Path
from typing import ClassVar, cast

from textual.binding import BindingType
from textual.widgets.option_list import Option

from tofuref.config import config
from tofuref.data.cache import get_cached_providers
from tofuref.data.helpers import get_registry_api
from tofuref.data.providers import Provider
from tofuref.widgets import keybindings
from tofuref.widgets.menu_option_list_base import MenuOptionListBase

LOGGER = logging.getLogger(__name__)


class ProvidersOptionList(MenuOptionListBase):
    BINDINGS: ClassVar[list[BindingType]] = [*MenuOptionListBase.BINDINGS, keybindings.OPEN_GITHUB, keybindings.GITHUB_STATS]

    def __init__(self, **kwargs):
        super().__init__(
            name="Providers",
            id="nav-provider",
            classes="nav-selector",
            **kwargs,
        )
        self.display = False
        self.fallback_providers_file = Path(__file__).resolve().parent.parent / "fallback" / "providers.json"

    def populate(
        self,
        providers: Collection[Provider] | None = None,
    ) -> None:
        if providers is None:
            providers = self.app.providers.values()
        self.clear_options()
        self.add_options(providers)
        self.border_subtitle = f"{len(providers):n} / {len(self.app.providers):n}"

    async def load_index(self) -> dict[str, Provider]:
        LOGGER.debug("Loading providers")

        data = await get_registry_api("index.json")
        await self.app.force_draw(initial=True)
        if not data:
            data = json.loads(self.fallback_providers_file.read_text())
            self.app.notify(
                "Something went wrong while fetching index of providers, using limited fallback.",
                title="Using fallback",
                severity="error",
            )

        LOGGER.debug("Got API response (or fallback)")
        await self.app.force_draw(initial=True)

        providers = await self.load_providers(data)

        providers = dict(sorted(providers.items(), key=lambda p: (p[1].bookmarked, p[1].cached, p[1].popularity), reverse=True))
        await self.app.force_draw(initial=True)
        return providers

    async def load_providers(self, data: dict) -> dict[str, Provider]:
        """
        Loads providers from API data. To show up in tofuref, it must:

        * have version
        * not be blocked in the registry
        * not be a fork
        * not be part of the organizations opentofu or terraform-providers, because those are just duplicates
        """
        providers = {}
        cached_providers = await get_cached_providers()

        for provider_json in data["providers"]:
            provider = Provider.from_json(provider_json)

            filter_in = (
                provider.versions,
                not provider.blocked,
                not provider.fork_of,
                provider.organization not in ["terraform-providers", "opentofu"],
            )
            if all(filter_in):
                providers[provider.display_name] = provider
                if self.app.bookmarks.check("providers", provider.identifying_name):
                    provider.bookmarked = True
                if provider.display_name in cached_providers:
                    provider.cached = True
        return providers

    async def on_option_selected(self, option: Option) -> None:
        __start_time = time.perf_counter()

        provider_selected = cast(Provider, option.prompt)
        self.app.active_provider = provider_selected

        await self.app.force_draw()
        await self.app.navigation_resources.load_provider_resources(provider_selected)

        self.app.navigation_resources.focus()

        await self.app.force_draw()
        self.replace_option_prompt_at_index(self.highlighted, option.prompt)

        if config.show_load_times:
            __load_time = time.perf_counter() - __start_time
            self.app.notify(f"Provider {provider_selected.name} loaded in {int(__load_time * 1000)}ms", timeout=10)

    def action_open_github(self):
        if self.highlighted is None:
            return
        option = self.get_option_at_index(self.highlighted)
        provider: Provider = option.prompt
        self.app.open_url(provider.github_url)

    async def action_github_stats(self):
        if self.highlighted is None:
            return
        option = self.get_option_at_index(self.highlighted)
        provider: Provider = option.prompt
        stats = await provider.github_stats()
        if stats:
            msg = f"Stars: [$primary]{stats['stars']}[/]\nOpen issues/PRs: [$primary]{stats['open_issues']}[/]"
            if stats["archived"]:
                msg += "\n[bold $warning]Archived[/]"
            self.app.notify(
                msg,
                title=f"GitHub stats for {provider.organization}/{provider.name}",
                timeout=15,
            )
        else:
            self.app.notify(
                "Something went wrong while fetching GitHub stats.",
                title="GitHub stats error",
                severity="error",
            )



from dataclasses import dataclass
from enum import Enum
from typing import TYPE_CHECKING, Literal

import frontmatter
from textual.content import Content

from tofuref.config import config
from tofuref.data import emojis
from tofuref.data.cache import clear_from_cache
from tofuref.data.helpers import (
    get_registry_api,
)
from tofuref.data.meta import Item

if TYPE_CHECKING:
    from tofuref.data.providers import Provider


class ResourceType(Enum):
    RESOURCE = "resource"
    DATASOURCE = "datasource"
    GUIDE = "guide"
    FUNCTION = "function"


@dataclass
class Resource(Item):
    name: str
    provider: "Provider"
    type: ResourceType
    _content: str | None = None
    cached: bool = False
    _title: str | None = None
    bookmarked: bool = False
    kind: Literal["resources"] = "resources"

    def __lt__(self, other: "Resource") -> bool:
        return self.name < other.name

    def __gt__(self, other: "Resource") -> bool:
        return self.name > other.name

    @property
    def display_name(self):
        return self._title if self._title is not None else self.name

    def visualize(self):
        cached_icon = emojis.CACHE if config.theme.emoji else "[$success]C[/] "
        bookmark_icon = emojis.BOOKMARK if config.theme.emoji else "[$success]B[/] "
        if self.bookmarked:
            prefix = bookmark_icon
        elif self.cached:
            prefix = cached_icon
        else:
            prefix = ""
        resource_icon = emojis.RESOURCE_TYPE[self.type.value] if config.theme.emoji else f"[$secondary]{self.type.value[0].upper()}[/]"
        return Content.from_markup(f"{resource_icon} {prefix}{self.display_name}")

    @property
    def identifying_name(self):
        return f"{self.provider.name}_{self.type.value}_{self.name}"

    def __hash__(self):
        return hash(self.identifying_name)

    @property
    def endpoint(self) -> str:
        return f"{self.provider.organization}/{self.provider.name}/{self.provider.active_version}/{self.type.value}s/{self.name}.md"

    async def content(self):
        if self._content is None:
            doc_data = await get_registry_api(self.endpoint, json=False)
            doc = frontmatter.loads(doc_data)
            self._content = doc.content
            if self.type == ResourceType.GUIDE:
                # noinspection PyTypeChecker
                self._title = doc.metadata["page_title"]
            self.cached = True
        return self._content

    async def clear_from_cache(self) -> None:
        # TODO clear all versions?
        if self.cached:
            await clear_from_cache(self.endpoint)
            self.cached = False



import asyncio
from typing import ClassVar, cast

from textual.binding import BindingType
from textual.widgets.option_list import Option

from tofuref.data.providers import Provider
from tofuref.data.resources import Resource
from tofuref.widgets.keybindings import BACK, LEFT_BACK
from tofuref.widgets.menu_option_list_base import MenuOptionListBase


class ResourcesOptionList(MenuOptionListBase):
    BINDINGS: ClassVar[list[BindingType]] = [*MenuOptionListBase.BINDINGS, BACK, LEFT_BACK]

    def __init__(self, **kwargs):
        super().__init__(
            name="Resources",
            id="nav-resources",
            classes="nav-selector",
            **kwargs,
        )
        self.display = False

    def populate(
        self,
        provider: Provider | None = None,
        resources: list[Resource] | None = None,
    ) -> None:
        self.clear_options()
        if provider is None:
            return

        if resources is None:
            self.add_options(provider.resources + provider.datasources)
        else:
            self.add_options(resources)

    async def load_provider_resources(
        self,
        provider: Provider,
    ):
        self.loading = True
        self.app.content_markdown.loading = True
        # Let the loading paint
        await self.app.force_draw()
        loaders = [self.render_overview(provider), provider.load_resources(bookmarks=self.app.bookmarks)]
        await asyncio.gather(*loaders)
        # Let the content update behind the loading screen
        await self.app.force_draw()
        self.app.content_markdown.loading = False
        self.populate(provider)
        self.focus()
        self.highlighted = 0
        self.loading = False

    async def render_overview(self, provider):
        overview = await provider.overview()
        await self.app.content_markdown.update(overview)
        await self.app.force_draw()

    async def on_option_selected(self, option: Option):
        resource_selected = cast(Resource, option.prompt)
        self.app.active_resource = resource_selected
        self.app.content_markdown.loading = True
        was_cached = resource_selected.cached

        content = await resource_selected.content()
        await self.app.content_markdown.update(content)
        is_cached = resource_selected.cached
        if was_cached != is_cached:
            self.replace_option_prompt_at_index(self.highlighted, option.prompt)
        self.app.content_markdown.document.focus()
        self.app.content_markdown.loading = False

    def action_back(self):
        self.app.action_providers()



from textual.app import ComposeResult
from textual.containers import Container
from textual.widgets import Label, Link


class StatusText(Label):
    def __init__(self, content: str, **kwargs) -> None:
        content = f"\\[ {content} ]"
        super().__init__(content, **kwargs)

    @property
    def content(self) -> str:
        return super().content

    @content.setter
    def content(self, value: str) -> None:
        super(Label, Label).content.__set__(self, f"\\[ {value} ]")


class Status(Container):
    DEFAULT_CSS = """
    Status {
        layout: horizontal;
        background: $surface;
        column-span: 1;
        height: 3;
        border: round $accent;
        border-right: none;
        border-left: none;
    }

    Status > #header-status {
        layout: grid;
        width: 100%;
        grid-size: 2;
        grid-columns: 1fr auto;
    }

    #header-info {
        width: auto;
        align: left middle;
        column-span: 1;
        layout: horizontal;
        padding: 0 1;
    }

    #header-links {
        width: auto;
        align: right middle;
        column-span: 1;
        layout: horizontal;
    }

    #header-links Link {
        padding: 0 1;
    }
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.provider = StatusText("-")
        self.version = StatusText("-")
        self.resource = StatusText("Welcome")
        self._path_label = Label("")

    def compose(self) -> ComposeResult:
        with Container(id="header-status"):
            with Container(id="header-info"):
                yield self.provider
                yield Label(" ⇢ ")
                yield self.version
                yield Label(" ⇢ ")
                yield self.resource
            with Container(id="header-links"):
                yield Link("GitHub", url="https://github.com/djetelina/tofuref")
                yield Link("Changelog", url="https://github.com/djetelina/tofuref/blob/main/CHANGELOG.md")



import json as jsonlib
import logging

import httpx

from tofuref import __version__
from tofuref.config import config
from tofuref.data.cache import cached_file_path, get_from_cache, save_to_cache

LOGGER = logging.getLogger(__name__)

CODEBLOCK_REGEX = r"^```([a-z]+)\n([\s\S]*?)^```"


async def get_registry_api(endpoint: str, json: bool = True) -> dict[str, dict] | str:
    """
    Sends GET request to opentofu providers registry to a given endpoint
    and returns the response either as a JSON or as a string. It also "logs" the request.

    Local cache is used to save/retrieve API responses.
    """
    uri = f"https://api.opentofu.org/registry/docs/providers/{endpoint}"
    if cached_content := await get_from_cache(endpoint):
        LOGGER.info(f"Using cached file for {endpoint} from {await cached_file_path(endpoint)}")
        return jsonlib.loads(cached_content) if json else cached_content
    LOGGER.info(f"Cache miss for {endpoint}")
    LOGGER.debug("Starting async client")
    async with httpx.AsyncClient(headers={"User-Agent": f"tofuref v{__version__}"}) as client:
        LOGGER.debug("Client started, sending request")
        try:
            r = await client.get(uri, timeout=config.http_request_timeout)
            LOGGER.debug("Request sent, response received")
        except Exception as e:
            LOGGER.error("Something went wrong", exc_info=e)
            return ""

    # Saving as text, because we are loading JSON if desired during cache hit
    LOGGER.info(f"Saving {endpoint} to cache")
    await save_to_cache(endpoint, r.text)

    return r.json() if json else r.text



from tofuref.data.resources import ResourceType
from tofuref.main import TofuRefApp, _parse_lookup_resource
from tofuref.startup import StartupTarget


def test_parse_lookup_resource():
    assert _parse_lookup_resource("github_repository") == ("github", "repository")
    assert _parse_lookup_resource("aws_instance") == ("aws", "instance")
    assert _parse_lookup_resource("repository") == (None, "repository")
    assert _parse_lookup_resource("aws") == (None, "aws")


def test_parse_lookup_resource_single_underscore():
    result = _parse_lookup_resource("test_resource")
    assert result == ("test", "resource")


async def test_cli_resource_auto_lookup(mock_cache_path, mock_http_requests, patch_bookmarks):
    app = TofuRefApp(startup=StartupTarget(resource="github_repository"))
    async with app.run_test() as pilot:
        await pilot.pause()
        assert app.navigation_providers.highlighted is not None


async def test_cli_provider_and_resource(mock_cache_path, mock_http_requests, patch_bookmarks):
    app = TofuRefApp(startup=StartupTarget(provider="integrations/github", resource="repository"))
    async with app.run_test() as pilot:
        await pilot.pause()
        assert app.active_provider is not None


async def test_cli_data_auto_lookup(mock_cache_path, mock_http_requests, patch_bookmarks):
    app = TofuRefApp(startup=StartupTarget(data="github_actions_environment_secrets"))
    async with app.run_test() as pilot:
        await pilot.pause()
        assert app.navigation_providers.highlighted is not None
        assert app.active_resource is not None
        assert app.active_resource.name == "actions_environment_secrets"
        assert app.active_resource.type == ResourceType.DATASOURCE


async def test_cli_provider_and_data(mock_cache_path, mock_http_requests, patch_bookmarks):
    app = TofuRefApp(startup=StartupTarget(provider="integrations/github", data="actions_environment_secrets"))
    async with app.run_test() as pilot:
        await pilot.pause()
        assert app.active_provider is not None



import re
from typing import ClassVar

from textual.binding import Binding, BindingType
from textual.widgets import MarkdownViewer, Tree

from tofuref.config import config
from tofuref.data.helpers import CODEBLOCK_REGEX
from tofuref.widgets import keybindings


class ContentWindow(MarkdownViewer):
    DEFAULT_CSS = """
    ContentWindow {
        border-left: $border-style $primary;
        border-right: $border-style $primary;
        scrollbar-color: $primary-darken-2;
        & > MarkdownTableOfContents {
            dock:right;
            border-left: $border-style $primary;
            border-right: $border-style $primary;
            color: $foreground;
            background: $surface;
            &:focus-within {
                color: $foreground;
                background: $surface;
                background-tint: $foreground 5%;
                border-left: $border-style $accent;
                border-right: $border-style $accent;
            }
        }
        &:focus-within {
            border-left: $border-style $accent;
            border-right: $border-style $accent;
            scrollbar-color: $primary;
        }
        & >MarkdownTableOfContents > Tree {
            background: $surface;
        }
    }
"""

    BINDINGS: ClassVar[list[BindingType]] = [
        Binding("k", "scroll_up", "Scroll Up", show=False),
        Binding("j", "scroll_down", "Scroll Down", show=False),
        Binding("ctrl+f", "page_up", "Page Up", show=False),
        Binding("ctrl+b", "page_down", "Page Down", show=False),
        Binding("G", "scroll_end", "Bottom", show=False),
        Binding("u", "yank", "Copy code blocks", show=False),
        Binding("y", "yank", "Copy code blocks"),
        Binding("t", "toggle_toc", "Toggle TOC"),
        Binding("B", "open_browser", "Open in browser"),
        keybindings.BACK,
        keybindings.LEFT_BACK,
    ]

    def __init__(self, content=None, **kwargs):
        welcome_content = f"""
# Welcome to tofuref!

Keyboard-first reference viewer for OpenTofu/Terraform providers and their resources.

Changelog: https://github.com/djetelina/tofuref/blob/main/CHANGELOG.md

---

## Quick start

1. Use `p`, `r` and `c` to navigate between three main windows.
2. Press `s` or `/` to search providers and resources.
3. Use arrow keys (or VIM `j`/`k`) to select a provider, then a resource.
4. View rich docs in this window. Press `t` to toggle the Table of Contents.
5. Press `y` (or `u`) to copy code blocks — pick the exact snippet you need.
6. Press `B` to open the current page in your browser.

Tip: Use `b` to bookmark frequently used items and prioritize them in future sort orders.

---

## Keyboard reference

### Actions
| keybindings | action |
|------|--------|
| `s`, `/` | **search** in the context of providers and resources |
| `u`, `y` | Context aware copying (using a provider/resource) |
| `v` | change active provider **version** |
| `b` | persistently bookmark an item to prioritize them in sorting when next re-ordered |
| `q`, `ctrl+q` | **quit** tofuref |
| `t` | toggle **table of contents** from content window |
| `B` | from content window, open active page in browser |
| `ctrl+l` | display **log** window |
| `ctrl+g` | open **GitHub** repository for provider |
| `ctrl+s` | Show **stats** of provider's github repo |

### Focus management

| keybindings | action |
|------|--------|
| `p` | focus **providers** window |
| `r` | focus **resources** window |
| `c` | focus **content** window |
| `backspace`, `left arrow` | focus parent window |
| `tab` | focus next window |
| `shift+tab` | focus previous window |

### In-window navigation

- Arrow keys, PageUp/PageDown, Home/End, or your mouse.
- VIM-style: `j`/`k` for line scroll, `ctrl+f` page down, `ctrl+b` page up, `G` to end.

---

## Get in touch
* GitHub: https://github.com/djetelina/tofuref"""

        self.content = content if content is not None else welcome_content
        super().__init__(
            self.content,
            show_table_of_contents=False,
            **kwargs,
        )

    async def update(self, markdown: str) -> None:
        markdown = sanitize_markdown(markdown).strip()

        markdown_stripped = try_to_strip_length(markdown)
        if markdown_stripped != markdown:
            self.app.notify(
                "Some sections were removed, open in browser ([bold]B[/]) for the full page.",
                title="Content was too long",
                severity="warning",
                timeout=30,
            )
            incomplete_infobox = """> The content of this document was too long, tofuref had to remove seme sections.
>
> To view the full page, open it in your browser (keybind `B`).
            """
            markdown = f"{incomplete_infobox}\n\n{markdown_stripped}"

        self.content = markdown
        await self.document.update(self.content)

    def action_toggle_toc(self):
        self.show_table_of_contents = not self.show_table_of_contents
        if self.show_table_of_contents:
            toc = self.table_of_contents.query_one(Tree)
            toc.focus()
            toc.action_cursor_down()
        else:
            self.document.focus(scroll_visible=False)

    def action_yank(self):
        code_blocks = re.findall(CODEBLOCK_REGEX, self.content, re.MULTILINE | re.DOTALL)
        if self.app.code_block_selector.has_parent:
            self.app.code_block_selector.parent.remove_children([self.app.code_block_selector])
        if not code_blocks:
            return
        self.parent.mount(self.app.code_block_selector)
        self.app.code_block_selector.set_new_options(code_blocks)
        self.display = False
        # self.screen.maximize(self.app.code_block_selector)

    def action_open_browser(self):
        if not self.app.active_provider:
            url = "https://github.com/djetelina/tofuref/blob/main/CHANGELOG.md"
        else:
            provider = self.app.active_provider.display_name
            resource_type = self.app.active_resource.type.value
            resource_name = self.app.active_resource.name
            url = f"https://search.opentofu.org/provider/{provider}/latest/docs/{resource_type}s/{resource_name}"
        self.app.open_url(url)

    def action_back(self):
        if self.app.active_resource:
            self.app.action_providers()
        self.app.action_resources()

    # Without this, the Markdown viewer would try to open a file on a disk, while the Markdown itself will open a browser link (desired)
    async def go(self, location):
        return None


def sanitize_markdown(markdown: str) -> str:
    """
    Place to sanitize content that is incompatible with textual's Markdown.
    """
    markdown = markdown.replace("–", "-").replace("—", "-")  # noqa: RUF001
    return re.sub(r"<.*?>", "", markdown)


def remove_sections_for_nested(markdown: str, min_dots: int) -> str:
    flags = re.M | re.I
    dot_lookahead = rf"(?=(?:[^.\n]*\.){{{min_dots},}})"

    pattern = rf"^\s{{0,3}}\#{{1,6}}\s(?=[^\n]*Nested){dot_lookahead}[^\n]*(?:\n(?!\s{{0,3}}\#{{1,6}}\s).*)*(?:\n)?"
    return re.sub(pattern, "", markdown, flags=flags)


def try_to_strip_length(markdown: str) -> str:
    for i in range(5, 0, -1):
        if len(markdown) < config.markdown_length_target:
            break
        markdown = remove_sections_for_nested(markdown, i)
    return markdown



import asyncio
import logging
import os
import shutil
import subprocess
from asyncio import create_subprocess_shell
from dataclasses import dataclass, field
from typing import Literal

import frontmatter
import httpx
from textual.content import Content

from tofuref import __version__
from tofuref.config import config
from tofuref.data import emojis
from tofuref.data.bookmarks import Bookmarks
from tofuref.data.cache import clear_from_cache, get_cached_resources
from tofuref.data.helpers import (
    get_registry_api,
)
from tofuref.data.meta import Item
from tofuref.data.resources import Resource, ResourceType

LOGGER = logging.getLogger(__name__)


@dataclass
class Provider(Item):
    organization: str
    name: str
    description: str
    fork_count: int
    blocked: bool
    popularity: int
    _overview: str | None = None
    _active_version: str | None = None
    versions: list[dict[str, str]] = field(default_factory=list)
    fork_of: str | None = None
    raw_json: dict | None = None
    resources: list[Resource] = field(default_factory=list)
    datasources: list[Resource] = field(default_factory=list)
    functions: list[Resource] = field(default_factory=list)
    guides: list[Resource] = field(default_factory=list)
    bookmarked: bool = False
    cached: bool = False
    kind: Literal["providers"] = "providers"
    _github_stats: dict[str, str] | None = None

    @classmethod
    def from_json(cls, data: dict) -> "Provider":
        return cls(
            organization=data["addr"]["namespace"],
            name=data["addr"]["name"],
            description=data["description"],
            fork_count=data["fork_count"],
            blocked=data["is_blocked"],
            popularity=data["popularity"],
            versions=data["versions"],
            fork_of=data.get("fork_of", {}).get("display"),
            raw_json=data,
        )

    @property
    def display_name(self) -> str:
        return f"{self.organization}/{self.name}"

    @property
    def identifying_name(self) -> str:
        return self.display_name

    @property
    def github_url(self) -> str:
        # It would be safer to get the url from registry, but for now let's assume this pattern will stick for opentofu too :)
        return f"https://github.com/{self.organization}/terraform-provider-{self.name}"

    @property
    def active_version(self) -> str:
        if self._active_version is None:
            self._active_version = self.versions[0]["id"]
        return self._active_version

    @active_version.setter
    def active_version(self, value: str) -> None:
        self._active_version = value
        self.resources = []
        self._overview = None

    @property
    def endpoint(self) -> str:
        return f"{self.organization}/{self.name}/{self.active_version}/index.md"

    def _endpoint_wildcard_version(self) -> str:
        return self.endpoint.replace(self.active_version, "*")

    @property
    def use_configuration(self) -> str:
        return f"""    {self.name} = {{
      source  = "{self.organization}/{self.name}"
      version = "{self.active_version.lstrip("v")}"
    }}"""

    async def overview(self) -> str:
        if self._overview is None:
            doc_data = await get_registry_api(self.endpoint, json=False)
            doc = frontmatter.loads(doc_data)
            self._overview = doc.content
            self.cached = True
        return self._overview

    async def load_resources(self, bookmarks: Bookmarks) -> None:
        if self.resources:
            self.sort_resources()
        await self.reload_resources(bookmarks)

    async def reload_resources(self, bookmarks: Bookmarks) -> None:
        self.resources = []
        resource_data = await get_registry_api(f"{self.organization}/{self.name}/{self.active_version}/index.json")
        for g in sorted(resource_data["docs"]["guides"], key=lambda x: x["name"]):
            self.resources.append(Resource(g["name"], self, type=ResourceType.GUIDE))
        for r in sorted(resource_data["docs"]["resources"], key=lambda x: x["name"]):
            self.resources.append(Resource(r["name"], self, type=ResourceType.RESOURCE))
        for d in sorted(resource_data["docs"]["datasources"], key=lambda x: x["name"]):
            self.resources.append(Resource(d["name"], self, type=ResourceType.DATASOURCE))
        for f in sorted(resource_data["docs"]["functions"], key=lambda x: x["name"]):
            self.resources.append(Resource(f["name"], self, type=ResourceType.FUNCTION))

        cached_resources = await get_cached_resources(self.organization, self.name, self.active_version)
        preload_content = []
        for resource in self.resources:
            if bookmarks.check("resources", resource.identifying_name):
                resource.bookmarked = True
            if f"{resource.type.value}s/{resource.name}" in cached_resources:
                resource.cached = True
            if resource.cached and resource.type == ResourceType.GUIDE:
                # Guide titles should override names taken from the filename
                # But we don't want to cache all of them just to get the titles
                # so we load them only if they are cached
                preload_content.append(resource.content())

        await asyncio.gather(*preload_content)

        self.sort_resources()

    def sort_resources(self) -> None:
        type_order = {ResourceType.GUIDE: 0, ResourceType.RESOURCE: 1, ResourceType.DATASOURCE: 2, ResourceType.FUNCTION: 3}

        self.resources.sort(key=lambda x: (-x.bookmarked, -x.cached, type_order[x.type], x.name))

    def visualize(self) -> Content:
        cached_icon = emojis.CACHE if config.theme.emoji else "[$success]C[/] "
        bookmark_icon = emojis.BOOKMARK if config.theme.emoji else "[$success]B[/] "
        if self.bookmarked:
            prefix = bookmark_icon
        elif self.cached:
            prefix = cached_icon
        else:
            prefix = ""
        return Content.from_markup(f"{prefix}[dim italic]{self.organization}[/]/{self.name}")

    async def clear_from_cache(self) -> None:
        if self.cached:
            await clear_from_cache(self._endpoint_wildcard_version())
            # Also delete overview
            await clear_from_cache(self._endpoint_wildcard_version().replace(".md", ".json"))
            self.cached = False

    async def github_stats(self):
        # Not the prettiest, but all the http requests and cache handling should be refactored soon
        if self._github_stats:
            return self._github_stats
        headers = {"User-Agent": f"tofuref v{__version__}", "X-GitHub-Api-Version": "2022-11-28", "Accept": "application/vnd.github+json"}
        if os.getenv("GITHUB_TOKEN"):
            LOGGER.info("GitHub token found in env")
            headers["Authorization"] = f"Bearer {os.getenv('GITHUB_TOKEN')}"
        elif shutil.which("gh"):
            process = await create_subprocess_shell("gh auth token", stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, _ = await process.communicate()
            token = stdout.decode().strip()
            # Ensure the stdout really produced token
            if token.startswith("gh"):
                LOGGER.info("Using gh to get GitHub token")
                headers["Authorization"] = f"Bearer {token}"

        async with httpx.AsyncClient(headers=headers) as client:
            try:
                ret = await client.get(
                    f"https://api.github.com/repos/{self.organization}/terraform-provider-{self.name}", timeout=config.http_request_timeout
                )
            except httpx.HTTPError as _:
                return self._github_stats
        data = ret.json()

        ok_status_code = 200
        if ret.status_code == ok_status_code:
            stats = {
                "stars": data.get("stargazers_count", 0),
                "open_issues": data.get("open_issues_count", 0),
                "archived": data.get("archived", False),
            }
            if stats["stars"] == 0:
                stats["stars"] = data.get("watchers_count", 0)
            # localize
            for k, v in stats.items():
                if not isinstance(v, bool):
                    stats[k] = f"{int(v):n}"

            self._github_stats = stats
        return self._github_stats



from datetime import datetime

from anyio import Path
from platformdirs import user_cache_path

from tofuref.config import config


def get_cache_path() -> Path:
    """Gets the user cache path converted to Path"""
    # This can't be a variable because pytest doesn't like that.
    # And we leave ensure_exists synchronous, because it doesn't matter really
    # (The real reason is it's also in the default factory of bookmarks, so who knows what gets called first)
    return Path(user_cache_path("tofuref", ensure_exists=True))


async def cached_file_path(endpoint: str, glob: bool = False) -> Path:
    """
    Args:
        endpoint: http endpoint of the registry API
        glob: Looks for glob matches in the cache directory, never use with saving into cache.

    Returns:
        Path to the cached file for a given endpoint.
        If glob is True, returns the first match, check `exists()`!.
    """
    filename = endpoint.replace("/", "_")
    if glob:
        matches = [p async for p in get_cache_path().glob(filename)]
        if matches:
            return Path(matches[0])
    return get_cache_path() / filename


async def save_to_cache(endpoint: str, contents: str) -> None:
    if not config.disable_cache:
        cached_file = await cached_file_path(endpoint)
        await cached_file.write_text(contents)


async def is_provider_index_expired(file: Path) -> bool:
    """
    Provider index is mutable, we consider it expired after 31 days (unconfigurable for now)

    One request per month is not too bad (we could have static fallback for the cases where this is hit when offline).
    New providers that people actually want probably won't be showing too often, so a month should be okay.
    """
    timeout = config.index_cache_duration_days * 86400
    now = datetime.now().timestamp()
    return file == await cached_file_path("index.json") and now - (await file.stat()).st_mtime >= timeout


async def get_from_cache(endpoint: str) -> str | None:
    """Loads from cache, unless the provider index is expired or cache is disabled in config"""
    cached_file = await cached_file_path(endpoint)
    if not await cached_file.exists() or await is_provider_index_expired(cached_file) or config.disable_cache:
        return None
    return await cached_file.read_text()


async def clear_from_cache(endpoint: str) -> None:
    """Removes all cached files (for all versions if given with a wildcard) for a given endpoint"""
    cached_file = await cached_file_path(endpoint, glob=True)
    while await cached_file.exists():
        await cached_file.unlink()
        # Load another
        cached_file = await cached_file_path(endpoint, glob=True)


async def get_cached_providers() -> list[str]:
    """For optimized marking of cached providers"""
    cached = [f.stem async for f in get_cache_path().glob("*_index.json")]
    return [f"{c.split('_')[0]}/{c.split('_')[1]}" for c in cached]


async def get_cached_resources(organization, name, version) -> list[str]:
    """For optimized marking of cached resources"""
    cached = [f.stem async for f in get_cache_path().glob(f"{organization}_{name}_{version}_*_*.md")]
    return [f"{c.split('_')[3]}/{'_'.join(c.split('_')[4:]).split('.')[0]}" for c in cached]



from collections.abc import Iterable
from dataclasses import dataclass

from tofuref.data.providers import Provider


@dataclass
class StartupTarget:
    provider: str | None = None
    resource: str | None = None
    data: str | None = None


def find_best_provider(prefix: str, providers: Iterable[Provider]) -> Provider | None:
    prefix = prefix.lower()
    matches = [p for p in providers if prefix in p.name.lower()]
    if not matches:
        return None
    matches.sort(key=lambda x: (-x.bookmarked, -x.cached, -x.popularity))
    return matches[0]



import asyncio
import locale
import logging
import sys
import time
from typing import ClassVar

import click
import httpx
from packaging.version import Version
from textual import on
from textual.app import App, ComposeResult
from textual.binding import Binding, BindingType
from textual.containers import Center, Container, Middle
from textual.fuzzy import Matcher
from textual.theme import BUILTIN_THEMES
from textual.widgets import (
    Footer,
    Input,
    OptionList,
    Select,
    TabbedContent,
    TabPane,
)

from tofuref import __version__
from tofuref.config import config
from tofuref.data.bookmarks import Bookmarks
from tofuref.data.providers import Provider
from tofuref.data.resources import ResourceType
from tofuref.startup import StartupTarget, find_best_provider
from tofuref.widgets import (
    CodeBlockSelect,
    ContentWindow,
    Logo,
    ProvidersOptionList,
    ResourcesOptionList,
    SearchInput,
    StartProgress,
    Status,
)

LOGGER = logging.getLogger(__name__)
locale.setlocale(locale.LC_ALL, "")


def _parse_lookup_resource(lookup: str) -> tuple[str | None, str | None]:
    if "_" not in lookup:
        return None, lookup

    prefix, resource = lookup.split("_", 1)
    return prefix, resource


async def resolve_lookup(app: "TofuRefApp", lookup: str, is_data: bool = False) -> None:
    provider_prefix, item_name = _parse_lookup_resource(lookup)
    if not provider_prefix:
        app.notify(f"Invalid lookup format: '{lookup}'. Use 'provider_resource' or 'provider_data' format.", severity="warning")
        return

    provider = find_best_provider(provider_prefix, app.providers.values())
    if provider is None:
        app.notify(f"No provider found matching '{provider_prefix}'", severity="warning")
        return

    await app._navigate_to_provider(provider.display_name)

    if item_name:
        item_name = item_name.lower()
        if is_data:
            await app._navigate_to_data(item_name)
        else:
            await app._navigate_to_resource(item_name)


class TofuRefApp(App):
    CSS_PATH = "tofuref.tcss"
    BINDINGS: ClassVar[list[BindingType]] = [
        ("/", "search", "Search"),
        Binding("s", "search", "Search", show=False),
        ("y", "use", "Use provider"),
        Binding("u", "use", "Use provider", show=False),
        ("v", "version", "Provider Version"),
        Binding("p", "providers", "Providers", show=False),
        Binding("r", "resources", "Resources", show=False),
        Binding("c", "content", "Content", show=False),
        Binding("ctrl+l", "log", "Show Log", show=False),
        Binding("q", "quit", "Quit"),
    ]
    TITLE = "TofuRef - OpenTofu Provider Reference"
    ESCAPE_TO_MINIMIZE = False

    def __init__(self, *args, startup: StartupTarget | None = None, **kwargs):
        self.startup = startup or StartupTarget()

        self.__start_time: float = time.perf_counter()
        # We are updating config in the tests, we need to reload config
        if "pytest" in sys.modules:
            config.load(reset=True)
        # We have to do this super early, otherwise tests are flaky
        for theme in BUILTIN_THEMES.values():
            theme.variables.update({"border-style": config.theme.borders_style})

        super().__init__(*args, **kwargs)
        # Widgets for easier reference, they could be replaced by query method
        self.content_markdown = ContentWindow()
        self.navigation_providers = ProvidersOptionList()
        self.navigation_resources = ResourcesOptionList()
        self.search = SearchInput()
        self.code_block_selector = CodeBlockSelect()
        self.initial_progress = StartProgress(total=7, show_eta=False, show_percentage=False)

        # Internal state
        self.bookmarks = Bookmarks()
        self.providers = {}
        self._active_provider = None
        self._active_resource = None
        self._search_target: ProvidersOptionList | ResourcesOptionList | None = None

        self.theme = config.theme.ui
        self.__load_time: float | None = None

    @property
    def active_provider(self):
        return self._active_provider

    @active_provider.setter
    def active_provider(self, provider):
        status = self.query_one("Status")
        status.provider.content = provider.display_name
        status.version.content = provider.active_version
        status.resource.content = "Overview"
        self._active_provider = provider

    @property
    def active_resource(self):
        return self._active_resource

    @active_resource.setter
    def active_resource(self, resource):
        self.query_one("Status").resource.content = f"{resource.provider.name}_{resource.name}"
        self._active_resource = resource

    def compose(self) -> ComposeResult:
        # Navigation
        with Container(id="header"):
            yield Status()
            yield Logo()

        with Container(id="navigation"):
            with Center(), Middle():
                yield self.initial_progress
            with TabbedContent():
                with TabPane("Providers", id="providers-tab"):
                    yield self.navigation_providers
                with TabPane("Resources", id="resources-tab"):
                    yield self.navigation_resources

        self.screen.maximize(self.initial_progress)

        # Main content area
        with Container(id="content"):
            yield self.content_markdown

        yield Footer()

    async def on_ready(self) -> None:
        # Draw the initial layout
        await self.force_draw(initial=True)
        self.call_next(self.load_content)
        self.call_later(self.check_for_new_version)

    async def load_content(self) -> None:
        await self.force_draw(initial=True)
        await self.load_providers_and_bookmarks()
        await self.force_draw(initial=True)
        self.navigation_providers.populate()
        await self.rearrange_loaded()
        await self.force_draw(initial=True)
        if config.show_load_times:
            self.__load_time = time.perf_counter() - self.__start_time
            self.notify(f"Loaded in {int(self.__load_time * 1000)}ms", timeout=10)

    async def load_providers_and_bookmarks(self) -> None:
        to_load = [self.navigation_providers.load_index(), self.bookmarks.async_post_init()]
        self.providers, _ = await asyncio.gather(*to_load)

    async def rearrange_loaded(self) -> None:
        # Start showing providers and resources
        self.navigation_providers.display = True
        self.navigation_resources.display = True
        self.screen.minimize()
        # Focus the first provider
        self.navigation_providers.focus()
        self.navigation_providers.highlighted = 0
        # We no longer need the progress bar
        await self.screen.remove_children([self.query_one("Center")])
        LOGGER.info("Initial load complete")

        target = self.startup
        if target.provider is not None:
            await self._navigate_to_provider(target.provider)
            if target.resource is not None:
                resource = target.resource
                if "_" in resource:
                    _, resource = resource.split("_", 1)
                await self._navigate_to_resource(resource)
            if target.data is not None:
                data = target.data
                if "_" in data:
                    _, data = data.split("_", 1)
                await self._navigate_to_data(data)
        else:
            if target.resource is not None:
                if "_" in target.resource:
                    await resolve_lookup(self, target.resource)
                else:
                    self.notify(
                        "No provider specified. Use 'provider_resource' format (e.g., 'github_repository') or provide --provider explicitly.",
                        severity="warning",
                    )
            if target.data is not None:
                if "_" in target.data:
                    await resolve_lookup(self, target.data, is_data=True)
                else:
                    self.notify(
                        "No provider specified. Use 'provider_data' format (e.g., 'github_actions_environment_secrets') "
                        "or provide --provider explicitly.",
                        severity="warning",
                    )

    async def force_draw(self, seconds=0.001, initial=False):
        """Used to yield event loop to textual so that it can render."""
        if initial:
            self.initial_progress.advance(1)
        if "pytest" not in sys.modules:
            await asyncio.sleep(seconds)

    async def check_for_new_version(self) -> None:
        newest_version = await get_current_pypi_version()
        version = Version(__version__)
        if version < newest_version:
            self.notify(f"✨ Version {newest_version} is available!\n[dim]Update now for the latest improvements[/dim]", timeout=20)

    def _focused_search_target(self) -> ProvidersOptionList | ResourcesOptionList | None:
        for searchable in (self.navigation_providers, self.navigation_resources):
            if searchable.has_focus:
                return searchable
        return None

    def _close_search(self) -> None:
        if self.search.has_parent:
            self.search.remove()
        self._search_target = None

    def action_search(self) -> None:
        """Focus the search input."""
        searchable = self._focused_search_target() or self._search_target
        if searchable is None:
            return

        self._close_search()
        self._search_target = searchable

        search_host = searchable.parent
        if not isinstance(search_host, TabPane):
            return

        self.search.value = ""
        self.search.styles.position = "absolute"
        self.search.styles.layer = "top"
        search_host.mount(self.search)
        self.search.offset = searchable.offset + (0, max(0, searchable.size.height - 3))  # noqa: RUF005
        self.search.focus()

    async def action_use(self) -> None:
        if not self.content_markdown.document.has_focus:
            if self.active_provider:
                to_copy = self.active_provider.use_configuration
            elif self.navigation_providers.highlighted is not None:
                highlighted_provider = self.navigation_providers.options[self.navigation_providers.highlighted].prompt
                to_copy = highlighted_provider.use_configuration
            else:
                return
            self.copy_to_clipboard(to_copy)
            self.notify(to_copy, title="Copied to clipboard", timeout=10)

    def action_providers(self) -> None:
        self.navigation_providers.focus()

    def action_resources(self) -> None:
        self.navigation_resources.focus()

    def action_content(self) -> None:
        self.content_markdown.document.focus()

    async def action_version(self) -> None:
        if self.active_provider is None:
            self.notify(
                "Provider Version can only be changed after one is selected.",
                title="No provider selected",
                severity="warning",
            )
            return
        if self.navigation_resources.children:
            await self.navigation_resources.remove_children("#version-select")
        else:
            version_select = Select.from_values(
                (v["id"] for v in self.active_provider.versions),
                prompt="Select Provider Version",
                allow_blank=False,
                value=self.active_provider.active_version,
                id="version-select",
            )
            await self.navigation_resources.mount(version_select)
            version_select.action_show_overlay()

    @on(Select.Changed, "#version-select")
    async def change_provider_version(self, event: Select.Changed) -> None:
        if event.value != self.active_provider.active_version:
            self.active_provider.active_version = event.value
            self.query_one("Status").version.content = event.value
            await self.navigation_resources.load_provider_resources(self.active_provider)
            # TODO open the resource that was selected before version change
            await self.navigation_resources.remove_children("#version-select")

    @on(Input.Changed, "#search")
    def search_input_changed(self, event: Input.Changed) -> None:
        query = event.value.strip()
        search_target = self._search_target
        if search_target == self.navigation_providers:
            if not query:
                self.navigation_providers.populate()
            elif config.fuzzy_search:
                matcher = Matcher(query)
                scored = [(v, matcher.match(p)) for p, v in self.providers.items()]
                self.navigation_providers.populate([v for v, s in sorted(scored, key=lambda x: x[1], reverse=True) if s > 0])
            else:
                self.navigation_providers.populate([v for p, v in self.providers.items() if query in p])
        elif search_target == self.navigation_resources and self.active_provider:
            if not query:
                self.navigation_resources.populate(
                    self.active_provider,
                )
            elif config.fuzzy_search:
                matcher = Matcher(query)
                all_items = self.active_provider.resources + self.active_provider.datasources
                scored = [(r, matcher.match(r.name)) for r in all_items]
                self.navigation_resources.populate(
                    self.active_provider,
                    [r for r, s in sorted(scored, key=lambda x: x[1], reverse=True) if s > 0],
                )
            else:
                self.navigation_resources.populate(
                    self.active_provider,
                    [r for r in (self.active_provider.resources + self.active_provider.datasources) if query in r.name],
                )

    @on(Input.Submitted, "#search")
    def search_input_submitted(self, _event: Input.Submitted) -> None:
        search_target = self._search_target
        if search_target is not None:
            search_target.focus()
            if search_target.option_count > 0:
                search_target.highlighted = 0
        self._close_search()

    @on(OptionList.OptionSelected)
    async def option_list_option_selected(self, event: OptionList.OptionSelected) -> None:
        await event.control.on_option_selected(event.option)

    async def _navigate_to_provider(self, provider_name: str) -> None:
        provider_name = provider_name.lower()
        for idx, provider in enumerate(self.navigation_providers.options):
            if provider.prompt.display_name.lower() == provider_name:
                self.navigation_providers.highlighted = idx
                option = self.navigation_providers.get_option_at_index(idx)
                await self.navigation_providers.on_option_selected(option)
                return
        self.notify(f"Provider '{provider_name}' not found", severity="warning")

    async def _navigate_to_item(self, item_name: str, item_type: str) -> None:
        item_name = item_name.lower()
        if self.active_provider is None:
            self.notify(f"No provider selected, cannot navigate to {item_type}", severity="warning")
            return

        self.query_one(TabbedContent).active = "resources-tab"
        self.navigation_resources.populate(self.active_provider)
        for idx, option in enumerate(self.navigation_resources.options):
            prompt = option.prompt
            if prompt.name.lower() == item_name and (
                prompt.type == ResourceType.DATASOURCE if item_type == "data" else prompt.type == ResourceType.RESOURCE
            ):
                self.navigation_resources.highlighted = idx
                await self.navigation_resources.on_option_selected(option)
                return
        self.notify(f"{item_type.capitalize()} '{item_name}' not found", severity="warning")

    async def _navigate_to_resource(self, resource_name: str) -> None:
        await self._navigate_to_item(resource_name, "resource")

    async def _navigate_to_data(self, data_name: str) -> None:
        await self._navigate_to_item(data_name, "data")


async def get_current_pypi_version() -> Version:
    async with httpx.AsyncClient(headers={"User-Agent": f"tofuref v{__version__}"}) as client:
        try:
            r = await client.get("https://pypi.org/pypi/tofuref/json", timeout=config.http_request_timeout)
        except Exception as _:
            return Version("0.0.0")
        return Version(r.json()["info"]["version"])


def main() -> None:
    @click.command()
    @click.option("-p", "--provider", help="Provider to open on startup (e.g., 'integrations/github')")
    @click.option(
        "-r",
        "--resource",
        help="Resource to open on startup. Use 'provider_resource' format (e.g., 'github_repository') "
        "to auto-resolve provider, or just 'resource_name' with --provider",
    )
    @click.option(
        "-d",
        "--data",
        help="Data source to open on startup. Use 'provider_data' format (e.g., 'github_actions_environment_secrets') "
        "to auto-resolve provider, or just 'data_source_name' with --provider",
    )
    def cli(provider: str | None, resource: str | None, data: str | None) -> None:
        target = StartupTarget(provider=provider, resource=resource, data=data)
        LOGGER.debug("Starting tofuref with %s", target)
        TofuRefApp(startup=target).run()

    cli()


if __name__ == "__main__":
    main()
